#NVDAObjects/excel.py
#A part of NonVisual Desktop Access (NVDA)
#Copyright (C) 2006-2007 NVDA Contributors <http://www.nvda-project.org/>
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.

from comtypes import COMError
import comtypes.automation
import wx
import oleacc
import textInfos
import colors
import eventHandler
import gui
import ui
import winUser
import controlTypes
from . import Window
from .. import NVDAObjectTextInfo
import scriptHandler

xlA1 = 1

class ExcelBase(Window):
	"""A base that all Excel NVDAObjects inherit from, which contains some useful methods."""

	@staticmethod
	def excelWindowObjectFromWindow(windowHandle):
		try:
			pDispatch=oleacc.AccessibleObjectFromWindow(windowHandle,winUser.OBJID_NATIVEOM,interface=comtypes.automation.IDispatch)
		except (COMError,WindowsError):
			return None
		return comtypes.client.dynamic.Dispatch(pDispatch)

	@staticmethod
	def getCellAddress(cell, external=False):
		return cell.Address(False, False, xlA1, external)

	def fireFocusOnSelection(self):
		selection=self.excelWindowObject.Selection
		if selection.Count>1:
			obj=ExcelSelection(windowHandle=self.windowHandle,excelWindowObject=self.excelWindowObject,excelRangeObject=selection)
		else:
			obj=ExcelCell(windowHandle=self.windowHandle,excelWindowObject=self.excelWindowObject,excelCellObject=selection)
		eventHandler.executeEvent("gainFocus",obj)

class Excel7Window(ExcelBase):
	"""An overlay class for Window for the EXCEL7 window class, which simply bounces focus to the active excel cell."""

	def _get_excelWindowObject(self):
		return self.excelWindowObjectFromWindow(self.windowHandle)

	def event_gainFocus(self):
		self.fireFocusOnSelection()

class ExcelWorksheet(ExcelBase):

	role=controlTypes.ROLE_TABLE
	def __init__(self,windowHandle=None,excelWindowObject=None,excelWorksheetObject=None):
		self.excelWindowObject=excelWindowObject
		self.excelWorksheetObject=excelWorksheetObject
		super(ExcelWorksheet,self).__init__(windowHandle=windowHandle)
		for gesture in self.__changeSelectionGestures:
			self.bindGesture(gesture, "changeSelection")

	def _get_name(self):
		return self.excelWorksheetObject.name

	def _isEqual(self, other):
		if not super(ExcelWorksheet, self)._isEqual(other):
			return False
		return self.excelWorksheetObject.index == other.excelWorksheetObject.index

	def _get_firstChild(self):
		cell=self.excelWorksheetObject.cells(1,1)
		return ExcelCell(windowHandle=self.windowHandle,excelWindowObject=self.excelWindowObject,excelCellObject=cell)
		
	def script_changeSelection(self,gesture):
		gesture.send()
		if scriptHandler.isScriptWaiting():
			# Prevent lag if keys are pressed rapidly.
			return
		self.fireFocusOnSelection()
	
	script_changeSelection.canPropagate=True

	__changeSelectionGestures = (
		"kb:tab",
		"kb:shift+tab",
		"kb:upArrow",
		"kb:downArrow",
		"kb:leftArrow",
		"kb:rightArrow",
		"kb:control+upArrow",
		"kb:control+downArrow",
		"kb:control+leftArrow",
		"kb:control+rightArrow",
		"kb:home",
		"kb:end",
		"kb:control+home",
		"kb:control+end",
		"kb:shift+upArrow",
		"kb:shift+downArrow",
		"kb:shift+leftArrow",
		"kb:shift+rightArrow",
		"kb:shift+control+upArrow",
		"kb:shift+control+downArrow",
		"kb:shift+control+leftArrow",
		"kb:shift+control+rightArrow",
		"kb:shift+home",
		"kb:shift+end",
		"kb:shift+control+home",
		"kb:shift+control+end",
		"kb:shift+space",
		"kb:control+space",
		"kb:control+pageUp",
		"kb:control+pageDown",
		"kb:control+v",
	)

class ExcelCellTextInfo(NVDAObjectTextInfo):

	def _getFormatFieldAndOffsets(self,offset,formatConfig,calculateOffsets=True):
		formatField=textInfos.FormatField()
		fontObj=self.obj.excelCellObject.font
		if formatConfig['reportFontName']:
			formatField['font-name']=fontObj.name
		if formatConfig['reportFontSize']:
			formatField['font-size']=str(fontObj.size)
		if formatConfig['reportFontAttributes']:
			formatField['bold']=fontObj.bold
			formatField['italic']=fontObj.italic
			formatField['underline']=fontObj.underline
		if formatConfig['reportColor']:
			try:
				formatField['color']=colors.RGB.fromCOLORREF(int(fontObj.color))
			except COMError:
				pass
			try:
				formatField['background-color']=colors.RGB.fromCOLORREF(int(self.obj.excelCellObject.interior.color))
			except COMError:
				pass
		return formatField,(self._startOffset,self._endOffset)

import speech
import config
class ExcelCell(ExcelBase):
	@classmethod
	def kwargsFromSuper(cls,kwargs,relation=None):
		windowHandle=kwargs['windowHandle']
		excelWindowObject=cls.excelWindowObjectFromWindow(windowHandle)
		if not excelWindowObject:
			return False
		if isinstance(relation,tuple):
			excelCellObject=excelWindowObject.rangeFromPoint(relation[0],relation[1])
		else:
			excelCellObject=excelWindowObject.ActiveCell
		if not excelCellObject:
			return False
		kwargs['excelWindowObject']=excelWindowObject
		kwargs['excelCellObject']=excelCellObject
		return True
	
	def __init__(self,windowHandle=None,excelWindowObject=None,excelCellObject=None):
		self.excelWindowObject=excelWindowObject
		self.excelCellObject=excelCellObject
		super(ExcelCell,self).__init__(windowHandle=windowHandle)

	role=controlTypes.ROLE_TABLECELL
	TextInfo=ExcelCellTextInfo
	
	def script_toggleExcelBack(self,gesture):
		config.conf["keyboard"]["SpeakCellPosition"]=True
		config.conf["keyboard"]["SpeakCellContent"]=True
			
	def _get_cellName(self, row, col):
		cell = self._get_cell(row, col)
		name = self.getCellAddress(cell)
		if config.conf["keyboard"]["SpeakCellPosition"]:
			return name
		else:
			return " "
			
	def _get_cellValue(self, row, col):
		cell = self._get_cell(row, col)
		value = cell.Text
		if config.conf["keyboard"]["SpeakCellContent"]:
			return value
		else:
			return " "
		
	def _get_cell(self, row, col):
		ws=self.excelCellObject.Worksheet
		try:
			cell = ws.cells(row,col)
		except COMError:
			speech.speakMessage("blank",index=None)
			return ""
		name = self.getCellAddress(cell)
		value = cell.Text
		return cell

	def script_toggleFirstFewRows(self,gesture):
		key = gesture.logIdentifier
		keycmd = key.split('+')
		cmd = keycmd[1]
		num = int(cmd)
		cell = self.excelCellObject
		try:
			colnum = cell.column
			rownum = cell.row
		except COMError:
			speech.speakMessage("com error",index=None)	
			return None
		name = self._get_cellName(int(rownum), num)
		value = self._get_cellValue(int(rownum), num)
		speech.speakMessage(name+" "+value,index=None)	
	
	def script_toggleFirstFewColumns(self,gesture):
		key = gesture.logIdentifier
		keycmd = key.split('+')
		cmd = keycmd[2]
		num = int(cmd)
		cell = self.excelCellObject
		try:
			colnum = cell.column
			rownum = cell.row
		except COMError:
			speech.speakMessage("com error",index=None)	
			return None
		name = self._get_cellName(num, int(colnum))
		value = self._get_cellValue(num, int(colnum))
		speech.speakMessage(name+" "+value,index=None)	
		
	def script_toggleExcelCell(self,gesture):
		if config.conf["keyboard"]["SpeakCellPosition"]:
			onOff=_("off")
			msg = "toggle excel position off"
			speech.speakMessage(msg)
			config.conf["keyboard"]["SpeakCellPosition"]=False
			config.conf["keyboard"]["SpeakCellContent"]=True
		else:
			onOff=_("on")
			msg = "toggle excel position on"
			speech.speakMessage(msg)
			config.conf["keyboard"]["SpeakCellPosition"]=True
			config.conf["keyboard"]["SpeakCellContent"]=True			
		speech.speakMessage(onOff)
			
	__gestures = {
		"kb:NVDA+0": "toggleExcelCell",
		"kb:alt+1": "toggleFirstFewRows",
		"kb:alt+2": "toggleFirstFewRows",
		"kb:alt+3": "toggleFirstFewRows",
		"kb:alt+4": "toggleFirstFewRows",
		"kb:alt+5": "toggleFirstFewRows",
		"kb:alt+control+1": "toggleFirstFewColumns",
		"kb:alt+control+2": "toggleFirstFewColumns",
		"kb:alt+control+3": "toggleFirstFewColumns",
		"kb:alt+control+4": "toggleFirstFewColumns",
		"kb:alt+control+5": "toggleFirstFewColumns",
		
	}

	def _isEqual(self,other):
		if not super(ExcelCell,self)._isEqual(other):
			return False
		thisAddr=self.getCellAddress(self.excelCellObject,True)
		try:
			otherAddr=self.getCellAddress(other.excelCellObject,True)
		except COMError:
			#When cutting and pasting the old selection can become broken
			return False
		return thisAddr==otherAddr
	
	def _get_name(self):	
		name = self.getCellAddress(self.excelCellObject)
		if config.conf["keyboard"]["SpeakCellPosition"]:
			return name
		else:
			return " "
			
	def _get_value(self):
		value = self.excelCellObject.Text
		if config.conf["keyboard"]["SpeakCellContent"]:
			return value
		else:
			return " "
		
	def _get_description(self):
		if self.excelCellObject.HasFormula:
			return _("has formula")
		else:
			return ""
			
	def _get_parent(self):
		worksheet=self.excelCellObject.Worksheet
		return ExcelWorksheet(windowHandle=self.windowHandle,excelWindowObject=self.excelWindowObject,excelWorksheetObject=worksheet)

	def _get_next(self):
		try:
			next=self.excelCellObject.next
		except COMError:
			next=None
		if next:
			return ExcelCell(windowHandle=self.windowHandle,excelWindowObject=self.excelWindowObject,excelCellObject=next)

	def _get_previous(self):
		try:
			previous=self.excelCellObject.previous
		except COMError:
			previous=None
		if previous:
			return ExcelCell(windowHandle=self.windowHandle,excelWindowObject=self.excelWindowObject,excelCellObject=previous)

class ExcelSelection(ExcelBase):
	
	role=controlTypes.ROLE_TABLECELL
	print("ExcelSelection(ExcelBase):")
	def __init__(self,windowHandle=None,excelWindowObject=None,excelRangeObject=None):
		self.excelWindowObject=excelWindowObject
		self.excelRangeObject=excelRangeObject
		super(ExcelSelection,self).__init__(windowHandle=windowHandle)

	def _get_states(self):
		states=super(ExcelSelection,self).states
		states.add(controlTypes.STATE_SELECTED)
		return states

	def _get_name(self):
		firstCell=self.excelRangeObject.Item(1)
		lastCell=self.excelRangeObject.Item(self.excelRangeObject.Count)
		return _("%s %s through %s %s")%(self.getCellAddress(firstCell),firstCell.Text,self.getCellAddress(lastCell),lastCell.Text)

	def _get_parent(self):
		worksheet=self.excelRangeObject.Worksheet
		return ExcelWorksheet(windowHandle=self.windowHandle,excelWindowObject=self.excelWindowObject,excelWorksheetObject=worksheet)

	#Its useful for an excel selection to be announced with reportSelection script
	def makeTextInfo(self,position):
		if position==textInfos.POSITION_SELECTION:
			position=textInfos.POSITION_ALL
		return super(ExcelSelection,self).makeTextInfo(position)
